/*
                                 SMARTSET v1.0

                             S - Y - N - T - A - X

 SMARTSET without any arguments will display the current status of SMARTDRV.

 SMARTSET followed by any invalid commandline argument will cause a syntax
 screen, resembling and briefer than the following information, to be displayed.

                     SMARTSET [F|R|D|E|+{kBytes}|-{kBytes}]
                  | = OR. Multiple arguments are not allowed.
        F = Flush Dirty Tracks             R = Flush and Reset Cache
        D = Disable (turn off) Cache       E = Enable (turn on) Cache
        + = Increase Cache by kBytes       - = Decrease Cache by kBytes
            Range = 16 - 9999                  Range = 16 - 9999

 kBytes are 16kByte multiples. The program will round down any amount that
 falls between 16kByte boundaries. Increases that would cause the current
 cache memory size to exceed the configured maximum will be truncated.
 Decreases that would cause the current cache memory size to fall below the
 configured minimum will be truncated.

 SMARTSET makes no distinction between upper and lower case letters used as
 commandline arguments.

 Examples:
 SMARTSET F
        Flushes all dirty tracks to disk.

 SMARTSET R
        Flushes and Resets the buffers.

 SMARTSET D
        Flushes and Resets the buffers then Disables SMARTDRV.

 SMARTSET E
        Enables SMARTDRV if it has been Disabled.

                                    SMARTDRV
        MINIMUM = 1024kBytes, MAXIMUM = 3072kBytes, SIZENOW = 2048kBytes

 SMARTSET +128
        Increase cache size by 128kBytes to 2176kBytes.

 SMARTSET -128
        Decrease cache size by 128kBytes to 1920kBytes.
                                           
 SMARTSET +129
        Increase cache size by 128kBytes to 2176kBytes.
        The 129 was rounded down to 128.

 SMARTSET +2048
        Increase cache size by 1024kBytes to 3072kBytes.
        The 2048 was truncated to 1024.

 SMARTSET -2048
        Decrease cache size by 1024kBytes to 1024kBytes.
        The 2048 was truncated to 1024.

                    ***** DUE TO A BUG IN SMARTDRV.SYS *****
    If the cache minimum size is less than the amount of memory required to
 cache one track and the current cache size is below the one track minimum,
 the minimum increase is the number of kBytes required to bring the current
 size equal to or greater than the memory required for one track. SMARTSET
 forces a 32kByte minimum size if the current size falls below 32kBytes.
 Increases must make the current size greater than 32kBytes.

    If you have a hard drive that requires more than 32kBytes to cache 1
 track, you should enforce you own higher minimum.

    When you setup SMARTDRV.SYS, it would be wise to never set the minimum
 cache size below the number of bytes required to cache one track rounded up
 to the nearest 16kByte increment. SMARTDRV.SYS sets the cache size required
 for one track to the largest track if you have multiple hard drives.

 Refer to your dos manual for proper syntax for installing SMARTDRV.SYS.

 Examples:
                               HARDRIVE - 1 Track Buffer
    512 bytes/sector * 17 sectors/track = 8,704 = 16,384 = 16kBytes minimum

                                    SMARTDRV
           MINIMUM = 0kBytes, MAXIMUM = 2048Bytes, SIZENOW = 0kBytes

 SMARTSET +16
        This would produce an error as a 32kByte Minimum is forced.
 SMARTSET +32
        This would produce an error as a 32kByte Minimum is forced.
 SMARTSET +48
        This would increase the cache size by 48kBytes to 48kBytes.

                                    SMARTDRV
           MINIMUM = 0kBytes, MAXIMUM = 2048Bytes, SIZENOW = 16kBytes

 SMARTSET +16
        This would produce an error as a 32kByte Minimum is forced.
 SMARTSET +32
        This would increase the cache size by 32kBytes to 48kBytes.

 Dated:  January 25, 1992
 Author: Everett B. Delano
         P.O. Box 205
         Elk City, Ok. 73648-0205
*/

#pragma inline

#include <stdlib.h>
#include <dos.h>
#include <fcntl.h>
#include <conio.h>
#include <stdio.h>
#include <bios.h>


struct SD_READ  {
  char      unused_1;
  char      unused_2;
  char      IsActive;
  char      MemoryType;
  unsigned  Ticks;
  char      IsLocked;
  char      FlushOnReboot;
  char      unused_3;
  char      DoubleBuffer;
  void      far *OrgInt13;
  char      MinorVersion;
  char      MajorVersion;
  char      unused_4;
  char      unused_5;
  unsigned  SectorsRead;
  unsigned  SectorsHit;
  unsigned  SectorsBuffered;
  char      HitRatio;
  char      BufferRatio;
  unsigned  TracksInCache;
  unsigned  CurrentTracks;
  unsigned  LockedTracks;
  unsigned  DirtyTracks;
  unsigned  CurrentSize;
  unsigned  ConfiguredSize;
  unsigned  MinimumSize;
  char      far *GlobalLockFlag;
};

struct SD_WRITE	{
  char    command;
  union   {
    char  subcommand;
    int   size;
    char  far *address;
  };
};

#ifdef __cplusplus
  #define __CPPARGS ...
#else
  #define __CPPARGS
#endif

#define	AND	&&
#define	NOT	!
#define OR	||

unsigned	cdecl _dos_gethandlestatus ( int, unsigned * );
unsigned	cdecl _dos_ioctlread ( int, void far *, unsigned, unsigned * );
unsigned	cdecl _dos_ioctlwrite ( int, void far *, unsigned, unsigned * );

void    main ( int, char ** );

void    get_configuration ( void );
void    show_configuration ( void );
void    flush_cache ( void );
void    reset_cache ( void );
void    cache_off ( void );
void    cache_on ( void );
void    get_incdec ( char ** );
void    decrement_cache ( void );
void    increment_cache ( void );

void    set_traps ( void );
void    quit ( char * );
void    cleanup ( void );
void    interrupt CtrlC_trap ( __CPPARGS );
void    far critfail ( unsigned deverr, unsigned errval, unsigned far *devhdr );

struct SD_READ  sd_read;
struct SD_WRITE sd_write;

char      CmdLine [ 21 ];
char      SiZe [ 18 ];
char      syntax = 0;
char      terminating = 0;
int       sd_handle = -1;
int       IncDec;

unsigned  min;
unsigned  max;
unsigned  now;
unsigned  inc;

unsigned cdecl _dos_gethandlestatus ( int handle, unsigned *status )
{
  asm {
    mov ax,4400h
    mov bx,handle
    int 21h
    jc  done
    mov bx,status
    mov [bx],ax
    xor ax,ax
  }
done:
  return ( _AX );
}

unsigned cdecl _dos_ioctlread ( int handle, void far *buffer,
								unsigned count, unsigned *numread )
{
  asm {
    push ds
    mov ax,4402h
    mov bx,handle
    mov cx,count
    lds dx,buffer
    int 21h
    pop ds
    jc  done
    mov bx,numread
    mov [bx],ax
    xor ax,ax
  }
done:
  return ( _AX );
}

unsigned cdecl _dos_ioctlwrite ( int handle, void far *buffer,
								unsigned count, unsigned *numwrt )
{
  asm {
    push ds
    mov ax,4403h
    mov bx,handle
    mov cx,count
    lds dx,buffer
    int 21h
    pop ds
    jc  done
    mov bx,numwrt
    mov [bx],ax
    xor ax,ax
  }
done:
  return ( _AX );
}

void main ( register int argc, register char *argv [] )
{
  get_configuration ();
  sprintf ( SiZe, "       SIZE      " );

  if ( argc != 2 )
  {
    if ( sd_handle != -1 )
      _dos_close ( sd_handle );
    if ( argc > 1 )
    {
      syntax++;
      sprintf ( CmdLine, "<TOO MANY ARGS>" );
    }
    else
      sprintf ( CmdLine, "<NONE>" );

    show_configuration ();
    return;
  }

  min = sd_read.MinimumSize;
  max = sd_read.ConfiguredSize;
  now = sd_read.CurrentSize;

  set_traps ();
  ++argv;
  sprintf ( CmdLine, "%.20s", *argv );
  switch ( **argv )
  {
    case 'f':
    case 'F':
      sprintf ( CmdLine, "<FLUSH>" );
      flush_cache ();
      break;
    case 'r':
    case 'R':
      sprintf ( CmdLine, "<RESET>" );
      reset_cache ();
      break;
    case 'd':
    case 'D':
      sprintf ( CmdLine, "<DISABLE>" );
      cache_off ();
      break;
    case 'e':
    case 'E':
      sprintf ( CmdLine, "<ENABLE>" );
      cache_on ();
      break;
    case '+':
      get_incdec ( argv );
      sprintf ( CmdLine, "<INCREASE %4dkB>", IncDec );
      increment_cache ();
      break;
    case '-':
      get_incdec ( argv );
      sprintf ( CmdLine, "<DECREASE %4dkB>", IncDec );
      decrement_cache ();
      break;
    default:
      syntax++;
      break;
  }

  get_configuration ();
  show_configuration ();

  exit ( 0 );
}

void get_configuration ( void )
{
  register unsigned exitcode = 0;
  unsigned status;
  unsigned count;

  if ( sd_handle == -1 )
  {
    exitcode = _dos_open ( "SMARTAAR", O_RDWR, &sd_handle );

    if ( NOT exitcode )
    {
      exitcode = _dos_gethandlestatus ( sd_handle, &status );
      if ( NOT exitcode )
      {
        if ( ( status & 0x4080 ) != 0x4080 )
          exitcode = 0xFFFF;
      }
    }

    if ( exitcode )
      quit ( "Cannot open SMARTDRV.SYS device" );
  }

  exitcode = _dos_ioctlread ( sd_handle, &sd_read, sizeof ( sd_read ), &count );
  if ( exitcode OR count != sizeof ( sd_read ) )
    quit ( "Cannot read data from SMARTDRV.SYS device" );
}

void show_configuration ( void )
{
if ( syntax )
printf("\a\
\n͸\
\n SMARTSET Version 1.0                           COMMAND:%-20.20s \
\n                            S - Y - N - T - A - X                            \
\n                    SMARTSET [F|R|D|E|+{kBytes}|-{kBytes}]                   \
\n                  | = OR. Multiple arguments are not allowed.                \
\n         F = Flush Dirty Tracks             R = Flush and Reset Cache        \
\n         D = Disable (turn off) Cache       E = Enable (turn on) Cache       \
\n         + = Increase Cache by kBytes       - = Decrease Cache by kBytes     \
\n             Range = 16 - 9999                  Range = 16 - 9999            \
\n kBytes are 16kByte multiples. Program rounds down. -<Min|+>Max is Truncated \
\n Ex: SMARTSET +129 would increase cache size by 128 kBytes.                  \
\nĴ\
\n    If the cache minimum size is less than the amount of memory required to  \
\n cache one track and the current cache size is below the one track minimum,  \
\n the minimum increase is the number of kBytes required to bring the current  \
\n size equal to or greater than the memory required for one track. SMARTSET   \
\n forces a 32kByte minimum size if the current size falls below 32kBytes.     \
\n Increases must make the current size greater than 32kBytes.                 \
\n    When you setup SMARTDRV.SYS, it would be wise to never set the minimum   \
\n cache size below the number of bytes required to cache one track rounded up \
\n to the nearest 16kByte increment. SMARTDRV.SYS sets the cache size required \
\n for one track to the largest track if you have multiple hard drives.        \
\n Ex: 512 bytes/sector * 17 sectors/track = 8,704 = 16,384 = 16kBytes minimum \
\n;",
  CmdLine );
else
{
printf("\
\n͸\
\n SMARTSET Version 1.0    HELP = SMARTSET ?    COMMAND:%-20.20s \
\nĴ",
  CmdLine );
printf("\
\n                   SMARTDRV.SYS Version: %02u.%02u                               ",
	sd_read.MajorVersion, sd_read.MinorVersion );
printf("\
\n                                Enabled: %-3s                                 ",
	( sd_read.IsActive == 1 ) ? "Yes" : "No" );
printf("\
\n                            Memory Type: %-8s  <EXTended,EXPanded>       ",
	sd_read.MemoryType == 1 ? "EXTended" : "EXPanded" );
printf("\
\n                          Locked Tracks: %-3s / %-4u                          ",
	sd_read.IsLocked == 1 ? "Yes" : "No", sd_read.LockedTracks );
printf("\
\n                  Virtual DMA Buffering: %-8s  <Disabled,Forced,Dynamic> ",
	sd_read.DoubleBuffer == 0 ? "Disabled" :
	( sd_read.DoubleBuffer == 1 ? "Forced" : "Dynamic" ) );
printf("\
\n                        Flush on Reboot: %-3s                                 ",
  sd_read.FlushOnReboot == 1 ? "Yes" : "No" );

printf("\
\nĴ\
\n%17s      TRACKS              SECTORS            HIT RATE    \
\nĴ\
\n   Max: %5uk       Total: %4u      Attempted: %5u       Cache: %3u%%  ",
  SiZe,
	sd_read.ConfiguredSize << 4, sd_read.TracksInCache,
	sd_read.SectorsRead, sd_read.HitRatio );

printf("\
\n   Min: %5uk      In Use: %4u            Hit: %5u      Buffer: %3u%%  ",
	sd_read.MinimumSize << 4, sd_read.CurrentTracks,
	sd_read.SectorsHit, sd_read.BufferRatio );

printf("\
\n   Now: %5uk       Dirty: %4u       Buffered: %5u                    \
\n;",
	sd_read.CurrentSize << 4, sd_read.DirtyTracks, sd_read.SectorsBuffered );
}
}

void flush_cache ( void )
{
  unsigned exitcode, count;

  sd_write.command = 0x00;
  exitcode = _dos_ioctlwrite ( sd_handle, &sd_write, 1, &count );
  if ( exitcode OR count != 1 )
    quit ( "Cannot flush SMARTDRV.SYS cache" );
}

void reset_cache ( void )
{
  unsigned exitcode, count;

  sd_write.command = 0x01;
  exitcode = _dos_ioctlwrite ( sd_handle, &sd_write, 1, &count );
  if ( exitcode OR count != 1 )
    quit ( "Cannot reset SMARTDRV.SYS cache" );
}

void cache_off ( void )
{
  unsigned exitcode, count;

  sd_write.command = 0x02;
  exitcode = _dos_ioctlwrite ( sd_handle, &sd_write, 1, &count );
  if ( exitcode OR count != 1 )
    quit ( "Cannot deactivate SMARTDRV.SYS cache" );
}

void cache_on ( void )
{
  unsigned exitcode, count;

  sd_write.command = 0x03;
  exitcode = _dos_ioctlwrite ( sd_handle, &sd_write, 1, &count );
  if ( exitcode OR count != 1 )
    quit ( "Cannot activate SMARTDRV.SYS cache" );
}

void get_incdec ( char *argv [] )
{
  IncDec = atoi ( *argv + 1 );
  if ( IncDec < 16 || IncDec > 9999 )
    inc = 0u;
  else
    inc = ( unsigned )( IncDec >> 4 );
  if ( NOT inc )
  {
    syntax++;
    quit ( "" );
  }
}

void decrement_cache ( void )
{
  unsigned exitcode, count;

  if ( now > min )
  {
    if ( ( min + inc ) > now )
      inc = now - min;
    sd_write.command = 0x0B;
    sd_write.size = inc;
    exitcode = _dos_ioctlwrite ( sd_handle, &sd_write, 3, &count );
    if ( exitcode OR count != 3 )
      quit ( "Cannot decrease SMARTDRV.SYS cache" );
  }
  else
    inc = 0u;
  sprintf ( SiZe, " SIZE %04u-%04uk ", now << 4, inc << 4 );
}

void increment_cache ( void )
{
  unsigned exitcode, count;

  if ( now < max )
  {
    if ( ( now < 2 ) AND ( ( now + inc ) < 3 ) )
      quit ( "Increase parameter rejected to avoid SMARTDRV.SYS bug" );
    if ( ( now + inc ) > max )
      inc = max - now;
    sd_write.command = 0x0C;
    sd_write.size = inc;
    exitcode = _dos_ioctlwrite ( sd_handle, &sd_write, 3, &count );
    if ( exitcode OR count != 3 )
      quit ( "Cannot increase SMARTDRV.SYS cache" );
  }
  else
    inc = 0u;
  sprintf ( SiZe, " SIZE %04u+%04uk ", now << 4, inc << 4 );
}

void set_traps ( void )
{
  atexit ( cleanup );
  _dos_setvect ( 0x23, CtrlC_trap );
  _harderr ( critfail );
}

void cleanup ( void )
{
  terminating = 1;
  if ( sd_handle != -1 )
  {
    _dos_close ( sd_handle );
    sd_handle = -1;
  }
}

void interrupt CtrlC_trap ( __CPPARGS )
{
  if ( NOT terminating )
  {
    terminating++;
    quit ( "Terminated by user" );
  }
}

#pragma warn -par

void far critfail ( unsigned deverr, unsigned errval, unsigned far *devhdr )
{
  _hardresume ( _HARDERR_FAIL );
}

#pragma warn +par

void quit ( char *errmsg )
{
  if ( syntax )
     show_configuration ();
  else
    printf ( "\n\aSMARTSET ERROR: %s!", errmsg );
  exit ( 0xFF );
}
